/* jshint undef: true */
/* global plugin, require, com, org, java, sync, importPackage, Packages */
/* global Flag */

var facade = require("facade");
var Objects = require("convenience").Objects;
var Errors = require("convenience").Errors;
var _ = require("lodash");
var on = require("on")(plugin);
var FileUtil = Packages.jodd.io.FileUtil;
var StringUtilities = com.cattura.cattura_commons.utilities.StringUtilities;
var Executors = java.util.concurrent.Executors;
var MoreExecutors = com.google.common.util.concurrent.MoreExecutors;
var mediaProxy = facade.retrieveProxy("MediaProcessingProxy");
var environment = facade.retrieveProxy("EnvironmentProxy");
var devices = facade.retrieveProxy("DeviceProxy");
var PluginNames = com.cattura.agents.configuration_module.model.properties.PluginNames;
var Element = com.cattura.share.processor.data_provider.Element;
var Processor = com.cattura.share.processor.Processor;
var System = java.lang.System;
var File = java.io.File;
var LOGGER = plugin.info().logger;
var PPM_DIR = new File(environment.getStorageDirPath(), "ppm");
var TemplateHelper = com.cattura.agents.configuration_module.model.vos.capture.helpers.TemplateHelper;
var MediaTypes = com.cattura.cattura_media_commons.properties.MediaTypes;
var Runner = com.cattura.cattura_commons.vos.concurrent.Runner;
var IncidentReport = com.cattura.agents.configuration_module.model.vos.IncidentReport;
var FutureStatusWatcher = com.cattura.agents.configuration_module.model.vos.watchers.FutureStatusWatcher;
var Duration = java.time.Duration;
var ChronoUnit = java.time.temporal.ChronoUnit;
var Futures = com.google.common.util.concurrent.Futures;
var POLL_TIMEOUT_MS = 500; // Amount of time to await until dispatch new PPM data over event source [Milliseconds]
var WATCHER_TIMEOUT_SECONDS = 30; // Amount of time to wait for start/stop PPM session to complete, before sending an incident report [Seconds]

importPackage(com.cattura.share.processor.data_provider.elements);

/*
 * Clear out PPM graphs directory
 */
if (PPM_DIR.exists())
{
    FileUtil.cleanDir(PPM_DIR);
}

var PPM = (function() {
    var AUDIO_LEVEL_PLUGIN = "com.cattura.plugins.audio.analysis.level.AudioLevelPlugin";
    var ON_AUDIO_LEVEL_UPDATE_PLUGIN_SERVICE = "onAudioLevelUpdate";
    var EXECUTOR = MoreExecutors.listeningDecorator(Executors.newCachedThreadPool());
    var PPM_ONLY_CONFIG = {
        calculatePPM: new Flag(true),
        calculateMaxPeak: new Flag(false),
        calculateMinPeak: new Flag(false),
        calculateNoiseFloor: new Flag(false),
        calculateRMS: new Flag(false)
    };
    
    var _devices = []; // Detected devices.
    var _futureDevicePPMSessions = {}; // Device ID is used for key
    var _activeCaptureGraphs = {}; // MediaPackage ID is used for key
    var _ppmLevelsForGraph = {}; // Graph Root (as String) is used for key
    
    /*
     * Utility function to get PPM levels, for all devices, from a given graph
     * 
     * @param {type} graph
     * @returns {main_L40.ppmLevelsForGraph|main_L40.getPPMLevelsFromGraph.mainAnonym$0}
     */
    function getPPMLevelsFromGraph ($graph)
    {
        return _ppmLevelsForGraph[$graph.root.toString()] || {};
    }
    
    function updatePPMLevelsForGraph ($graph, $pluginResults, $outputID)
    {
        var graphRootAsStr = $graph.root.toString(),
            ppmLevelsForGraph = (_ppmLevelsForGraph[graphRootAsStr] = _ppmLevelsForGraph[graphRootAsStr] || {});
        
        $pluginResults = Objects.toJS($pluginResults, true);
        for (var key in $pluginResults)
        {
            var audioMetrics = $pluginResults[key];
            key = $outputID || key;
            ppmLevelsForGraph[key] = audioMetrics.DB;
        }
    }
    
    function getAudioLevelsForCaptureSessions ()
    {
        var allAudioLevels = {};
        for (var key in _activeCaptureGraphs)
        {
            var graph = _activeCaptureGraphs[key];
            if (!graph)
            {
                LOGGER.warning("Graph for " + key + " is not valid! (WTF?)");
                continue;
            }
            
            var audioLevelsForGraph = getPPMLevelsFromGraph(graph);
            var hasOneAudioLevel = _(audioLevelsForGraph).size() === 1;
            allAudioLevels[key] = hasOneAudioLevel ? _(audioLevelsForGraph).values().value()[0] : audioLevelsForGraph;
        }
        return allAudioLevels;
    }

    function setDevices ($someDevices)
    {
        _devices = Objects.toJS($someDevices);
    }

    function reset ()
    {
        _devices = [];
        _futureDevicePPMSessions = {};
        _activeCaptureGraphs = {};
        _ppmLevelsForGraph = {};
    }
    
    function setUpPPMForCaptureSessionGraph ($graph)
    {
        _(Objects.toJS($graph.nodes())).each(function ($node) {
            if (!(($node.type() === Processor.TYPE.WRITE || $node.type() === Processor.TYPE.INTEGRATION))) return;
            var outputID = $node.pluginID() + ":" + $node.id();
            
            if ($node.config().containsKey("id"))
            {
                outputID = JSON.parse($node.config().get("id").getValue().toString());
            }
            
            _(Objects.toJS($node.parents())).each(function ($parentNode) {
                addPPMToNode($graph, $parentNode, outputID);
            });
        });
        
        return $graph;
    }
    
    function addPPMToNode ($graph, $node, $outputID)
    {
        var module = mediaProxy.getCaptureModule();
        var plugin = module.getPlugin($node.pluginID());
        
        if (!plugin.getOutputInterests().contains(Element.TYPE.AUDIO_FORMAT.name()) || 
            !$node.output(Element.TYPE.AUDIO_FORMAT).isPresent())
        {
            LOGGER.info("Not hooking up node #" + $node.id() + " ( a " + $node.pluginID() + ") because it's not producing an audio output.");
            return;
        }
        
        /*
         * If PPM node is already attached to this audio node, just
         * reuse the existing PPM node.
         */
        var ppmNode = _(Objects.toJS($node.children())).find(function ($node) {
            return $node.pluginID().equals(AUDIO_LEVEL_PLUGIN);
        });
        
        if (!ppmNode)
        {
            ppmNode = $graph.add(AUDIO_LEVEL_PLUGIN)
                            ['linkParents(java.lang.Iterable)']([$node])
                            .configure(PPM_ONLY_CONFIG);
        }
        
        ppmNode.invokeService(ON_AUDIO_LEVEL_UPDATE_PLUGIN_SERVICE, {
            callback: Objects.toCallback(function($data) {
                updatePPMLevelsForGraph($graph, $data, $outputID);
            })
        });
    }
    
    function addCaptureSession ($req, $graph, $stopPPMSessionsForCaptureDevices)
    {
        setUpPPMForCaptureSessionGraph($graph);
        _activeCaptureGraphs[$req.mediaPackageID] = $graph;
        
        if ($stopPPMSessionsForCaptureDevices)
        {
            var helper = new TemplateHelper($graph);
            Futures.allAsList(_(Objects.toJS(helper.getSources())).filter(function ($source) {
                return $source.type === MediaTypes.Device.AUDIO;
            }).map(function ($source) {
                LOGGER.info("Stopping PPM monitoring session for [" + $source.sourceId + "] because a new capture session is starting.");
                return removeDevice($source.sourceId);
            }).value()).get();
        }
        
        return $graph;
    }
    
    function addDevice ($device)
    {
        if (!($device.name in _futureDevicePPMSessions))
        {
            try
            {
                LOGGER.info("PPM monitoring session for [" + $device.name + "] is starting up.");
                var ppmRoot = new File(PPM_DIR, System.currentTimeMillis() + "-" + StringUtilities.createSafeID($device.name));
                var mediaProcessingModule = mediaProxy.getCaptureModule();
                var ppmMonitorGraph = mediaProcessingModule.getGraphBuilder().get(ppmRoot);
                ppmMonitorGraph.make();

                var audioNode = ppmMonitorGraph.add($device.pluginID)
                                                ['linkParents(java.lang.Iterable)']([ppmMonitorGraph.node(0).get()])
                                                .configure({ device: $device.createConfiguration() });

                addPPMToNode(ppmMonitorGraph, audioNode);
                
                var futureSession = ppmMonitorGraph.startCapture();
                futureSession.addListener(function() {
                    var session = futureSession.get();
                    LOGGER.info("PPM monitoring session for [" + $device.name + "] started up.");
                    
                    session.onSessionTerminated(Objects.toCallback(function ($session) {
                        LOGGER.info("PPM monitoring session for [" + $device.name + "] terminated.");
                        
                        if (ppmRoot.exists())
                        {
                            try
                            {
                                FileUtil.deleteDir(ppmRoot);
                            }
                            catch (e)
                            {
                                LOGGER.info("Failed to delete PPM session directory: " + e);
                            }
                        }
                    }));
                }, EXECUTOR);
                
                /*
                 * Send an incident report if the PPM takes longer than expected to start up
                 */
                EXECUTOR.submit(new FutureStatusWatcher(futureSession, Duration.of(WATCHER_TIMEOUT_SECONDS, ChronoUnit.SECONDS), 5, Objects.toRunnable(function () {
                    var incident = new IncidentReport("PPM monitoring session for device [" + $device.name + "] is taking longer than expected to start up");
                    incident.setInternal(true);
                    facade.sendNotification(facade.HANDLE_EXCEPTION, incident);
                })));
                
                _futureDevicePPMSessions[$device.name] = futureSession;
            }
            catch (e)
            {
                facade.sendNotification(facade.HANDLE_EXCEPTION, new IncidentReport("PPM monitoring session for device [" + $device.name + "] failed to start up"));
            }
        }

        return _futureDevicePPMSessions[$device.name];        
    }
    
    function removeDevice ($device)
    {
        var deviceId = $device.name || $device;
        if (!(deviceId in _futureDevicePPMSessions))
        {
            return Futures.immediateFuture(false);
        }
        
        LOGGER.info("PPM monitoring session for [" + deviceId + "] is shutting down.");
        var future = EXECUTOR.submit(Objects.toCallable(function () {
            _futureDevicePPMSessions[deviceId].get().stop().get();
            delete _futureDevicePPMSessions[deviceId];
            return true;
        }));
        
        /*
         * Send an incident report if the PPM takes longer than expected to start up
         */
        EXECUTOR.submit(new FutureStatusWatcher(future, Duration.of(WATCHER_TIMEOUT_SECONDS, ChronoUnit.SECONDS), 5, Objects.toRunnable(function () {
            var incident = new IncidentReport("PPM monitoring session for device [" + deviceId + "] is taking longer than expected to terminate");
            incident.setInternal(true);
            facade.sendNotification(facade.HANDLE_EXCEPTION, incident);
        })));
        
        return future;
    }
    
    function addAllMonitorableDevices ()
    {
        var filteredMoniterableDevices = _(Objects.toJS(_devices)).filter(function(device) {
            if (!device.isUsable()) return false; // Filter unusable devices
            if (device.isUSBDevice()) return false; // Filter USB devices
            if (device.name.startsWith("bmd:")) return false; // Filter BMD devices (for now)
            
            return true;
        }).value();
        
        _(filteredMoniterableDevices).each(function($device) {
            var graph = addDevice($device);
            if (!graph)
            {
                LOGGER.warning("PPM graph for " + $device + " is not valid! (WTF?)");
                return;
            }
        });
    }
    
    function getAudioLevels ()
    {
        var levels = _({}).extend(getAudioLevelsForCaptureSessions()).value();
        
        _(getActiveDevicePPMSessions()).each(function($session) {
            _(levels).extend(getPPMLevelsFromGraph($session.graph)).value();
        });
        return levels;
    }
    
    function getActiveDevicePPMSessions ()
    {
        return _(_futureDevicePPMSessions).filter(function ($session) {
            return $session.isDone();
        }).map(function ($session) { 
            return $session.get();
        }).value();
    }
    
    return {
        getAudioLevels: getAudioLevels,
        setDevices: setDevices,
        addCaptureSession: addCaptureSession,
        addDevice: addDevice,
        removeDevice: removeDevice,
        addAllMonitorableDevices: addAllMonitorableDevices,
        reset: reset
    };
})();

on({
    AudioDevicesDetectedEvent: function($event)
    {
        if ($event.isFinishedDetecting())
        {
            PPM.setDevices($event.devices);
        }
    },

    MediaProcessingModuleLaunchEvent: function($event)
    {
        PPM.reset();
    },

    WebServerEvent: function($event)
    {
        var webServer = $event.getWebServer();
        var EventSourceHandler = org.webbitserver.EventSourceHandler;
        var EventSourceMessage = org.webbitserver.EventSourceMessage;
        
        /*
         * Register endpoint to handle all PPM metrics requests
         */
        webServer.add("/audiolevel", (function() {
            /*
             * Keep track of all the connections
             */
            var _connections = [];
            
            function hasConnections () {return _connections.length > 0;}
            
            /*
             * Runner to poll the PPM for data
             */
            var runner = new Runner.Atomic("PPM poller", true);
            runner.register(Objects.toRunnerListener(function () {
                /*
                 * Return immediately if there are no more connections.
                 */
                if (hasConnections())
                {
                    /*
                     * Add all monitorable devices to the PPM meter.
                     */
                    PPM.addAllMonitorableDevices();

                    /*
                     * Ask audio devices for their sound levels
                     */
                    var levels = PPM.getAudioLevels();
                    //                        if (_.isEmpty(levels)) continue; // Don't transmit if we don't have levels (ie no devices). Currently doesn't work b/c of FrameBuffer IOBuffer read error (b/c empty). Will have to investigate.

                    var jsonLevels = JSON.stringify(levels);
                    var message = new EventSourceMessage().event("levels").data(jsonLevels);
                    
                    /* jshint loopfunc:true */
                    _(_connections).each(function (connection) {
                        connection.send(message);
                    });
                    /* jshint loopfunc:false */
                }

                /*
                 * Sleep for a bit
                 */
                java.lang.Thread.sleep(POLL_TIMEOUT_MS);
            }));
            
            var startPolling = sync(function () {
                if (!this.isRunningOrWaiting())
                {
                    this.requestStartup();
                }
            }).bind(runner);
            
            /*
             * Basic event source handler
             */
            return new EventSourceHandler({
                onOpen: function ($connection)
                {
                    _connections.push($connection);
                    
                    startPolling();
                },
                onClose: function ($connection)
                {
                    _connections = _(_connections).without($connection).value();
                }
            });
        })());
    },
    
    /*
     * Invoked when a new CaptureRequest is made (ie user start capture or scheduled event).
     * Will not fire for PPM or live device preview session for feedback.
     */
    CaptureSessionConstructionEvent: function($event)
    {
        var req = $event.getRequest();
        var graph = $event.getGraph();
        PPM.addCaptureSession(req, graph, true);
    }
});

/*
 * If audio devices have been detected, then populate the PPM.
 */
if (devices.getDetectionProgress() >= 1)
{
    PPM.setDevices(devices.getAudioDeviceList());
}